@@ -3,6 +3,8 @@ from typing import Any, Callable, Optional, Sequence, Type, TypeVar, Union
 
 import torch
 import torch.nn.functional as F
+import xformers
+import xformers.ops
 from einops import pack, rearrange, reduce, repeat, unpack
 from torch import Tensor, einsum, nn
 from typing_extensions import TypeGuard
@@ -182,29 +184,64 @@ def ResnetBlock(
     return Module([conv_block, conv], lambda x: conv_block(x) + conv(x))
 
 
+class GRN(nn.Module):
+    """GRN (Global Response Normalization) layer from ConvNextV2 generic to any dim"""
+
+    def __init__(self, dim: int, channels: int):
+        super().__init__()
+        ones = (1,) * dim
+        self.gamma = nn.Parameter(torch.zeros(1, channels, *ones))
+        self.beta = nn.Parameter(torch.zeros(1, channels, *ones))
+        self.norm_dims = [d + 2 for d in range(dim)]
+
+    def forward(self, x: Tensor) -> Tensor:
+        Gx = torch.norm(x, p=2, dim=self.norm_dims, keepdim=True)
+        Nx = Gx / (Gx.mean(dim=1, keepdim=True) + 1e-6)
+        return self.gamma * (x * Nx) + self.beta + x
+
+
+def ConvNextV2Block(dim: int, channels: int) -> nn.Module:
+    block = nn.Sequential(
+        # Depthwise and LayerNorm
+        Conv(
+            dim=dim,
+            in_channels=channels,
+            out_channels=channels,
+            kernel_size=7,
+            padding=3,
+            groups=channels,
+        ),
+        nn.GroupNorm(num_groups=1, num_channels=channels),
+        # Pointwise expand
+        Conv(dim=dim, in_channels=channels, out_channels=channels * 4, kernel_size=1),
+        # Activation and GRN
+        nn.GELU(),
+        GRN(dim=dim, channels=channels * 4),
+        # Pointwise contract
+        Conv(
+            dim=dim,
+            in_channels=channels * 4,
+            out_channels=channels,
+            kernel_size=1,
+        ),
+    )
+
+    return Module([block], lambda x: x + block(x))
+
+
 def AttentionBase(features: int, head_features: int, num_heads: int) -> nn.Module:
-    scale = head_features**-0.5
     mid_features = head_features * num_heads
     to_out = nn.Linear(in_features=mid_features, out_features=features, bias=False)
 
     def forward(
         q: Tensor, k: Tensor, v: Tensor, mask: Optional[Tensor] = None
     ) -> Tensor:
-        h = num_heads
-        # Split heads
-        q, k, v = map(lambda t: rearrange(t, "b n (h d) -> b h n d", h=h), (q, k, v))
-        # Compute similarity matrix and add eventual mask
-        sim = einsum("... n d, ... m d -> ... n m", q, k) * scale
-        # Get attention matrix with softmax
-        attn = sim.softmax(dim=-1)
-        # Compute values
-        out = einsum("... n m, ... m d -> ... n d", attn, v)
-        out = rearrange(out, "b h n d -> b n (h d)")
+        # Use memory efficient attention
+        out = xformers.ops.memory_efficient_attention(q, k, v)
         return to_out(out)
 
     return Module([to_out], forward)
 
-
 def LinearAttentionBase(features: int, head_features: int, num_heads: int) -> nn.Module:
     scale = head_features**-0.5
     mid_features = head_features * num_heads
@@ -225,7 +262,6 @@ def LinearAttentionBase(features: int, head_features: int, num_heads: int) -> nn
 
     return Module([to_out], forward)
 
-
 def FixedEmbedding(max_length: int, features: int):
     embedding = nn.Embedding(max_length, features)
 
